package app

import (
	"context"
	"errors"
	"fmt"
	"gitea.pena/PenaSide/trashlog/clients"
	"gitea.pena/PenaSide/trashlog/controllers/http/health_check"
	"gitea.pena/PenaSide/trashlog/dal/clickhouse"
	trashlogProto "gitea.pena/PenaSide/trashlog/proto/generated"
	"gitea.pena/PenaSide/trashlog/server"
	"gitea.pena/PenaSide/trashlog/sink"
	"gitea.pena/PenaSide/trashlog/version"
	"github.com/skeris/appInit"
	"gitea.pena/PenaSide/hlog"
	"go.uber.org/zap"
	"google.golang.org/grpc"
	"net"
)

type App struct {
	err    chan error
	logger *zap.Logger
}

var (
	errInvalidOptions = errors.New("invalid options")
)

type Options struct {
	Development  bool   `env:"DEVELOPMENT" default:"true"`
	AppName      string `env:"APP_NAME" default:"trashlog"`
	GrpcURL      string `env:"GRPC_URL" default:":7113"`
	AdminHttpURL string `env:"ADMIN_HTTP_URL" default:"localhost:8000"`

	ClickhouseURL           string `env:"CLICKHOUSE_URL" default:"tcp://127.0.0.1:9000?debug=true"`
	DBPath                  string `env:"DB_PATH" default:"./recover.bolt"`
	ClickhouseTableMainName string `env:"CLICKHOUSE_TABLE_MAIN_NAME" default:"statistics"`

	TelegramToken     string `env:"TELEGRAM_TOKEN" default:"6712573453:AAFqTOsgwe_j48ZQ1GzWKQDT5Nwr-SAWjz8"`
	TelegramChannelID uint64 `env:"TELEGRAM_CHANNEL_ID" default:"1002217604546"`
}

var _ appInit.CommonApp = (*App)(nil)

type InfoSvcStarted struct{}
type InfoSvcReady struct{}
type InfoSvcShutdown struct {
	Signal string
}
type ErrorCanNotServe struct {
	Err error
}
type ErrorFailToConnectTarantool struct {
	Err error
}
type ErrorListener struct {
	Err error
}
type ErrorServe struct {
	Err error
}
type ErrorStoreInit struct {
	Err error
}
type ErrorStoreConnect struct {
	Err         error
	Credentials string
}

var zapOptions = []zap.Option{
	zap.AddCaller(),
	zap.AddCallerSkip(2),
	zap.AddStacktrace(zap.ErrorLevel),
}

func New(ctx context.Context, opts interface{}, ver appInit.Version) (appInit.CommonApp, error) {
	var (
		err     error
		errChan = make(chan error)
		logger  *zap.Logger
		options Options
		ok      bool
	)

	if options, ok = opts.(Options); !ok {
		return App{}, errInvalidOptions
	}

	if options.Development {
		logger, err = zap.NewDevelopment(zapOptions...)
	} else {
		logger, err = zap.NewProduction(zapOptions...)
	}
	if err != nil {
		return nil, err
	}

	logger = logger.Named(options.AppName)

	logger = logger.With(
		zap.String("SvcCommit", version.Commit),
		zap.String("SvcVersion", version.Release),
		zap.String("SvcBuildTime", version.BuildTime),
	)

	log := hlog.New(logger)
	log.Emit(InfoSvcStarted{})

	log.Emit(InfoSvcReady{})

	listener, err := net.Listen("tcp", options.GrpcURL)

	if err != nil {
		log.Emit(ErrorListener{Err: err})
		return nil, err
	}

	logStore, err := clickhouse.New(ctx, clickhouse.Deps{
		Log:                    log,
		Credentials:            options.ClickhouseURL,
		DBPath:                 options.DBPath,
		DBClickHouseName:       options.ClickhouseTableMainName,
		DBBufferClickHouseName: fmt.Sprintf("buffer_%s", options.ClickhouseTableMainName),
	})
	if err != nil {
		log.Emit(ErrorStoreConnect{
			Err:         err,
			Credentials: options.ClickhouseURL,
		})
		return nil, err
	}

	if err := logStore.Init(ctx); err != nil {
		log.Emit(ErrorStoreInit{
			Err: err,
		})
		return nil, err
	}

	trashlogClient, err := clients.NewTrashLogClient(options.GrpcURL)
	if err != nil {
		return nil, err
	}

	healthController := health_check.NewHealthCheck(health_check.Deps{
		Store:           logStore,
		Logger:          log,
		NameTable:       options.ClickhouseTableMainName,
		NameBufferTable: fmt.Sprintf("buffer_%s", options.ClickhouseTableMainName),
		TrashlogClient:  trashlogClient,
	})

	srv := server.NewServer(server.ServerConfig{
		Logger:      logger,
		Controllers: []server.Controller{healthController},
	})

	sinkSvc := sink.New(log, logStore)

	var grpcOpts []grpc.ServerOption
	grpcServer := grpc.NewServer(grpcOpts...)

	trashlogProto.RegisterTrashlogServer(grpcServer, sinkSvc)
	go func() {
		if err := grpcServer.Serve(listener); err != nil {
			log.Emit(ErrorServe{
				Err: err,
			})
			errChan <- err
		}
	}()

	go func() {
		if err := srv.Start(options.AdminHttpURL); err != nil {
			log.Emit(ErrorServe{
				Err: err,
			})
			errChan <- err
		}
	}()

	return App{
		err:    errChan,
		logger: logger,
	}, nil
}

func (a App) GetLogger() *zap.Logger {
	return a.logger
}

func (a App) GetErr() chan error {
	return a.err
}
