package health_check

import (
	"gitea.pena/PenaSide/trashlog/clients"
	"gitea.pena/PenaSide/trashlog/dal/clickhouse"
	"github.com/gofiber/fiber/v2"
	"gitea.pena/PenaSide/hlog"
)

type HealthCheck struct {
	store           *clickhouse.DAL
	logger          hlog.Logger
	nameTable       string
	nameBufferTable string
	trashlogClient  *clients.TrashLogClient
}

type Deps struct {
	Store           *clickhouse.DAL
	Logger          hlog.Logger
	NameTable       string
	NameBufferTable string
	TrashlogClient  *clients.TrashLogClient
}

func NewHealthCheck(deps Deps) *HealthCheck {
	return &HealthCheck{
		store:           deps.Store,
		logger:          deps.Logger,
		nameTable:       deps.NameTable,
		nameBufferTable: deps.NameBufferTable,
		trashlogClient:  deps.TrashlogClient,
	}
}

type ErrorPingEvent struct {
	Error error `json:"error"`
}

type ErrorCheckTableExistEvent struct {
	Error error `json:"error"`
}

type ErrorTableNotExistEvent struct {
	Error string `json:"error"`
}

// Liveness - простая проба, возвращает статус 200, если сервис жив.
func (receiver *HealthCheck) Liveness(ctx *fiber.Ctx) error {
	return ctx.SendStatus(fiber.StatusOK)
}

// TODO: Пересмотреть необходимость пинга базы как критерия readiness в будущем, особенно при использовании для автоскейлинга в Kubernetes.
// На данный момент пинг базы служит для мониторинга и оповещения админов при недоступности. Однако, при внедрении автоскейлинга
// зависимость от базы данных в readiness может создать узкое место. Если база недоступна, увеличение числа сервисов проблему не решит,
// так как логи всё равно будут сохраняться в recovery и обработаны позже.

// Readiness - проба, проверяющая доступность базы и необходимые таблицы, а так же rpc сервер.
func (receiver *HealthCheck) Readiness(ctx *fiber.Ctx) error {
	if err := receiver.store.Ping(); err != nil {
		receiver.logger.Emit(ErrorPingEvent{Error: err})
		return ctx.Status(fiber.StatusServiceUnavailable).SendString("Database unavailable")
	}

	exists, err := receiver.store.TableExists(receiver.nameTable)
	if err != nil {
		receiver.logger.Emit(ErrorCheckTableExistEvent{Error: err})
		return ctx.Status(fiber.StatusInternalServerError).SendString("Failed to check main table existence")
	}
	if !exists {
		receiver.logger.Emit(ErrorTableNotExistEvent{Error: "Main table does not exist"})
		return ctx.Status(fiber.StatusServiceUnavailable).SendString("Main table does not exist")
	}

	exists, err = receiver.store.TableExists(receiver.nameBufferTable)
	if err != nil {
		receiver.logger.Emit(ErrorCheckTableExistEvent{Error: err})
		return ctx.Status(fiber.StatusInternalServerError).SendString("Failed to check buffer table existence")
	}
	if !exists {
		receiver.logger.Emit(ErrorTableNotExistEvent{Error: "Buffer table does not exist"})
		return ctx.Status(fiber.StatusServiceUnavailable).SendString("Buffer table does not exist")
	}

	if err := receiver.trashlogClient.Ping(ctx.Context()); err != nil {
		receiver.logger.Emit(ErrorPingEvent{Error: err})
		return ctx.Status(fiber.StatusServiceUnavailable).SendString("RPC server unavailable")
	}

	return ctx.SendStatus(fiber.StatusOK)
}
