package bbolt_test

import (
	"github.com/pioz/faker"
	"github.com/stretchr/testify/assert"
	"github.com/themakers/bdd"
	bolt "go.etcd.io/bbolt"
	"go.uber.org/zap"
	"os"
	"gitea.pena/PenaSide/trashlog/dal/bbolt"
	"gitea.pena/PenaSide/trashlog/model"
	"testing"
	"time"
)

func TestNew(t *testing.T) {
	faker.SetSeed(time.Now().Unix())

	bdd.Scenario(t, "bbolt dal", func(t *testing.T, runID string) {
		var (
			masterd      *bbolt.DAL
			recoverFname = faker.StringWithSize(5)
			records      []model.Record
		)

		bdd.Act(t, "dal creation", func() {
			bdd.Test(t, "positive creation of dal", func() {
				d, err := bbolt.New(recoverFname)
				assert.NoError(t, err)
				assert.NotNil(t, d)

				stat, err := os.Stat(recoverFname)
				assert.NoError(t, err)
				assert.Equal(t, int(bbolt.Filemode), int(stat.Mode()))
				masterd = d
			})
			bdd.Test(t, "negative creation of dal", func() {
				d, err := bbolt.New(recoverFname)
				assert.Error(t, err)
				assert.Nil(t, d)
			})
		})

		bdd.Act(t, "put some data to bolt", func() {
			// idk ways to crush store function besides concurrency troubles
			for i := 0; i < 100; i++ {
				records = append(records, model.Record{
					Level: faker.Pick(
						zap.ErrorLevel.String(),
						zap.DebugLevel.String(),
						zap.DPanicLevel.String(),
						zap.FatalLevel.String(),
						zap.InfoLevel.String(),
						zap.PanicLevel.String(),
						zap.WarnLevel.String()),
					TS:      uint64(time.Now().Unix()),
					Message: faker.String(),
					Module: []string{
						faker.AddressCity(),
						faker.AddressCity(),
					},
					Stacktrace: faker.Paragraph(),
					KeyFields: map[string]interface{}{
						faker.AddressCity(): faker.String(),
						faker.AddressCity(): faker.String(),
					},
					CtxFields: map[string]interface{}{
						faker.FemaleFirstName(): faker.String(),
						faker.FemaleFirstName(): faker.String(),
					},
					SvcBuildTime: uint64(time.Now().Unix()),
					SvcVersion:   faker.StringWithSize(5),
					SvcCommit:    faker.StringWithSize(5),
					SvcFile:      faker.String(),
					SvcLine:      uint64(faker.Int()),
				})
			}

			t.Parallel()
			for i, _ := range records {
				bdd.Test(t, "store it in parallel", func() {
					assert.NoError(t, masterd.PutRecord(records[i]))
				})
			}
		})

		bdd.Act(t, "check that we obtain all records that was stored", func() {
			bdd.Test(t, "get all records", func() {
				i := 0
				assert.NoError(t, masterd.YieldRecovery(func(record model.Record) error {
					assert.Equal(t, records[i], record)
					i++
					return nil
				}))
			})
			bdd.Test(t, "bucket is empty", func() {
				i := 0
				assert.NoError(t, masterd.YieldRecovery(func(record model.Record) error {
					assert.Equal(t, records[i], record)
					i++
					return nil
				}))
				assert.Zero(t, i)
			})
		})

		bdd.Act(t, "bbolt dal closing", func() {
			bdd.Test(t, "positive dal closing", func() {
				assert.NoError(t, masterd.Close())

				db, err := bolt.Open(recoverFname, bbolt.Filemode, &bolt.Options{
					Timeout:      5 * time.Second,
					NoGrowSync:   false,
					FreelistType: bolt.FreelistArrayType,
				})
				assert.NoError(t, err)

				tx, err := db.Begin(false)
				assert.NoError(t, err)

				assert.NotNil(t, tx.Bucket([]byte("recover")))

				assert.NoError(t, tx.Rollback())

				assert.NoError(t, db.Close())

				assert.NoError(t, os.Remove(recoverFname))
			})
		})
	})

}

//var _ = Describe("Bbolt", func() {
//var (
//	db *bbolt.DAL
//	err error
//)
//
//defer func() {
//	if db != nil {
//		db.Close()
//	}
//}()
//
//	Context("Creation", func() {
//		It("CreateDB", func() {
//			db, err = bbolt.New("test.bolt")
//			Expect(db).ToNot(BeNil())
//			Expect(err).To(BeNil())
//		})
//	})
//
//	Context("Put", func() {
//		It("by dal", func() {
//			Expect(db.PutRecord(model.Record{})).To(BeNil())
//		})
//	})
//
//	Context("Get", func() {
//		It("by dal", func() {
//			Expect(db.YieldRecovery(func(record model.Record) error {
//				fmt.Println(record.TS == 0 , "A")
//				if record.TS == 0 {
//					return fmt.Errorf("%s", "cock")
//				}
//				return nil
//			})).To(BeNil())
//		})
//	})
//})
