package bbolt

import (
	"bytes"
	"encoding/gob"
	"github.com/rs/xid"
	bolt "go.etcd.io/bbolt"
	"os"
	"gitea.pena/PenaSide/trashlog/model"
	"time"
)

type DAL struct {
	db         *bolt.DB
	bucketName string
}

const (
	BucketName = "recover"
	Filemode   = os.FileMode(0664)
)

func New(dbpath string) (*DAL, error) {
	db, err := bolt.Open(dbpath, Filemode, &bolt.Options{
		Timeout:      5 * time.Second,
		NoGrowSync:   false,
		FreelistType: bolt.FreelistArrayType,
	})
	if err != nil {
		return nil, err
	}

	err = db.Update(func(tx *bolt.Tx) error {
		_, err := tx.CreateBucketIfNotExists([]byte(BucketName))
		if err != nil {
			return err
		}
		return nil
	})
	if err != nil {
		return nil, err
	}

	return &DAL{
		db:         db,
		bucketName: BucketName,
	}, nil
}

func (d *DAL) Close() error {
	return d.db.Close()
}

func (d *DAL) PutRecord(record model.Record) error {
	return d.db.Update(func(tx *bolt.Tx) error {
		b := tx.Bucket([]byte(d.bucketName))

		var buf bytes.Buffer
		in := gob.NewEncoder(&buf)
		if err := in.Encode(record); err != nil {
			return err
		}

		if err := b.Put(xid.New().Bytes(), buf.Bytes()); err != nil {
			return err
		}

		return nil
	})
}

func (d *DAL) YieldRecovery(yield func(record model.Record) error) error {
	return d.db.Update(func(tx *bolt.Tx) error {
		b := tx.Bucket([]byte(d.bucketName))

		return b.ForEach(func(k, v []byte) error {

			buf := bytes.NewBuffer(v)
			out := gob.NewDecoder(buf)
			var record model.Record
			if err := out.Decode(&record); err != nil {
				return err
			}

			err := yield(record)

			if err != nil {
				return err
			}

			return b.Delete(k)
		})
	})
}

func (d *DAL) GetAllRecords() (map[string]model.Record, error) {
	records := make(map[string]model.Record)

	err := d.db.View(func(tx *bolt.Tx) error {
		b := tx.Bucket([]byte(d.bucketName))
		if b == nil {
			return nil
		}

		c := b.Cursor()

		for k, v := c.First(); k != nil; k, v = c.Next() {
			buf := bytes.NewBuffer(v)
			out := gob.NewDecoder(buf)
			var record model.Record
			if err := out.Decode(&record); err != nil {
				return err
			}
			records[string(k)] = record
		}

		return nil
	})

	if err != nil {
		return nil, err
	}

	return records, nil
}

func (d *DAL) DeleteRecordByKey(key []byte) error {
	return d.db.Update(func(tx *bolt.Tx) error {
		b := tx.Bucket([]byte(d.bucketName))
		if b == nil {
			return nil
		}

		return b.Delete(key)
	})
}
