package sink

import (
	"context"
	"fmt"

	"gitea.pena/PenaSide/hlog"
	"gitea.pena/PenaSide/trashlog/dal/clickhouse"
	"gitea.pena/PenaSide/trashlog/model"
	trashlogProto "gitea.pena/PenaSide/trashlog/proto/generated"
)

type Sink struct {
	logger hlog.Logger
	store  *clickhouse.DAL
	trashlogProto.UnimplementedTrashlogServer
}

type InfoMobileRec struct {
	Level      string
	TS         uint64
	Message    string
	Module     []string
	Stacktrace string

	KeyFields    map[string]interface{}
	CtxFields    map[string]interface{}
	SvcBuildTime uint64
	SvcVersion   string
	SvcCommit    string
	SvcFile      string
	SvcLine      uint64
}

func (s *Sink) TgLog(_ context.Context, record *trashlogProto.Record) (*trashlogProto.Dummy, error) {
	data := Dto2daoRecord(record)
	s.logger.Emit(InfoMobileRec{
		Level:        data.Level,
		TS:           data.TS,
		Message:      data.Message,
		Module:       data.Module,
		Stacktrace:   data.Stacktrace,
		KeyFields:    data.KeyFields,
		CtxFields:    data.CtxFields,
		SvcBuildTime: data.SvcBuildTime,
		SvcVersion:   data.SvcVersion,
		SvcCommit:    data.SvcCommit,
		SvcFile:      data.SvcFile,
		SvcLine:      data.SvcLine,
	})
	return &trashlogProto.Dummy{}, nil
}

func New(log hlog.Logger, store *clickhouse.DAL) *Sink {
	return &Sink{
		logger: log.Module("sink"),
		store:  store,
	}
}

func (s *Sink) GetFields(
	_ context.Context,
	_ *trashlogProto.Dummy,
) (*trashlogProto.Fields2Add, error) {

	var result trashlogProto.Fields2Add
	result.Fields = make(map[string]string)

	result.Fields = s.store.Schema

	return &result, nil
}

func (s *Sink) Modify(
	ctx context.Context,
	fields *trashlogProto.Fields2Add,
) (*trashlogProto.NotModified, error) {

	if err := s.store.AddColumn(ctx, fields.Fields); err != nil {
		return &trashlogProto.NotModified{Columns: fields.Fields}, err
	}

	return &trashlogProto.NotModified{}, nil
}

func (s *Sink) Valve(stream trashlogProto.Trashlog_ValveServer) error {
	ctxChan := stream.Context()
	dataChan := make(chan *trashlogProto.Record)

	go func() {
		for {
      select {
      case <-ctxChan.Done():
        return
      default:
			record, err := stream.Recv()
			if err != nil {
				fmt.Println("RECV ERR", err)
				return
			}
      
			dataChan <- record
    }
		}
	}()

	for {
		select {
		case <-ctxChan.Done():
			if err := stream.SendAndClose(&trashlogProto.Dummy{}); err != nil {
				return err
			}
			fmt.Println("CLOSEVALVE")
			return nil
		case record := <-dataChan:
			fmt.Println("DATATATATA", record)
			s.store.PutRecord(ctxChan, Dto2daoRecord(record))
		}
	}
}

func Dto2daoRecord(in *trashlogProto.Record) model.Record {
	keyFields := map[string]interface{}{}
	for key, value := range in.KeyFields {
		keyFields[key] = Dto2daoValue(value)
	}

	ctxFields := map[string]interface{}{}
	for key, value := range in.CtxFields {
		ctxFields[key] = Dto2daoValue(value)
	}

	return model.Record{
		Level:        in.Level,
		TS:           in.TS,
		Message:      in.Message,
		Module:       in.Module,
		Stacktrace:   in.Stacktrace,
		SvcBuildTime: in.SvcFields.BuildTime,
		SvcVersion:   in.SvcFields.Version,
		SvcCommit:    in.SvcFields.Commit,
		SvcFile:      in.SvcFields.File,
		SvcLine:      in.SvcFields.Line,
		KeyFields:    keyFields,
		CtxFields:    ctxFields,
	}
}

func Dto2daoValue(in *trashlogProto.Value) interface{} {
	switch in.Value.(type) {
	case *trashlogProto.Value_Str:
		return in.GetStr()
	case *trashlogProto.Value_Double:
		return in.GetDouble()
	case *trashlogProto.Value_Num:
		return in.GetNum()
	case *trashlogProto.Value_Flag:
		fmt.Println("Flaag", in.GetFlag())
		if in.GetFlag() {
			return uint8(1)
		}
		return uint8(0)
	}

	return nil
}

func (s *Sink) Ping(context.Context, *trashlogProto.Dummy) (*trashlogProto.Dummy, error) {
	return &trashlogProto.Dummy{}, nil
}
