package main

import (
	"context"
	"flag"
	"fmt"
	"go/ast"
	"go/parser"
	"go/scanner"
	"go/token"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
	"io"
	"os"
	pb "gitea.pena/PenaSide/trashlog/proto/generated"
	"strings"
)

type Field struct {
	Name string
	Type string
}

var host = flag.String("host", "", "specify host:port of trashlog to update")

// go run main.go --host localhost:7113 D:\Goproject\customer\internal\models\trashLog.go

func main() {
	flag.Parse()

	args := flag.Args()
	if len(args) == 0 {
		fmt.Println("No args")
		return
	}

	if host == nil {
		*host = "localhost:7113"
	}

	var opts []grpc.DialOption
	opts = append(opts, grpc.WithTransportCredentials(insecure.NewCredentials()))

	ctx := context.Background()

	conn, err := grpc.Dial(*host, opts...)
	if err != nil {
		panic(err)
	}
	fmt.Println(conn.GetState())

	//defer conn.Close()
	client := pb.NewTrashlogClient(conn)
	//fmt.Println(client.)

	serverFields, err := client.GetFields(ctx,
		&pb.Dummy{},
	)
	fmt.Println(conn.GetState())

	if err != nil {
		panic(err)
	}

	fields := make(map[string]Field)

	toModify := pb.Fields2Add{
		Fields: make(map[string]string),
	}

	for _, filename := range args {
		fset := token.NewFileSet()
		fd, err := os.Open(filename)
		if err != nil {
			panic(err)
		}

		stat, err := fd.Stat()
		if err != nil {
			panic(err)
		}

		file := fset.AddFile(filename, fset.Base(), int(stat.Size()))

		filetext, err := io.ReadAll(fd)
		if err != nil {
			panic(err)
		}
		fmt.Println("TEST", string(filetext))

		var sc scanner.Scanner
		sc.Init(file, filetext, nil, scanner.ScanComments)

		f, err := parser.ParseFile(fset, filename, filetext, 0)
		if err != nil {
			panic(err)
		}

		for _, decl := range f.Decls {
			switch decl.(type) {
			case *ast.GenDecl:
				genDecl := decl.(*ast.GenDecl)
				for _, spec := range genDecl.Specs {
					switch spec.(type) {
					case *ast.TypeSpec:
						typeSpec := spec.(*ast.TypeSpec)
						switch {
						case strings.HasPrefix(typeSpec.Name.String(), "Info"):
							workWithRecords(typeSpec, &fields)
						case strings.HasPrefix(typeSpec.Name.String(), "Error"):
							workWithRecords(typeSpec, &fields)
						case strings.HasPrefix(typeSpec.Name.String(), "Debug"):
							workWithRecords(typeSpec, &fields)
						case strings.HasPrefix(typeSpec.Name.String(), "Warn"):
							workWithRecords(typeSpec, &fields)
						}
					}
				}
			}
		}

		withFound := false

		ast.Inspect(f, func(node ast.Node) bool {
			switch node.(type) {
			case *ast.CallExpr:
				bl := node.(*ast.CallExpr)
				switch bl.Fun.(type) {
				case *ast.SelectorExpr:
					fun := bl.Fun.(*ast.SelectorExpr)
					if fun.Sel.Name == "With" &&
						strings.Contains(
							strings.ToLower(
								fmt.Sprint(
									fun.X,
								),
							),
							"log",
						) {
						for _, arg := range bl.Args {
							complit, ok := arg.(*ast.CompositeLit)
							if ok {
								for _, fld := range complit.Elts {
									d, ok := fld.(*ast.KeyValueExpr)
									if ok {

										fmt.Println("SSSSS", d.Key, d.Value)
									}
								}
							}

						}
						withFound = true
					}
				}
			case *ast.KeyValueExpr:
				bl := node.(*ast.KeyValueExpr)
				if withFound {
					//fmt.Println("SOOOOQA3", bl.Key, bl.Value)
					var fieldname string

					key, ok := bl.Key.(*ast.BasicLit)
					if !ok {
						break
					}

					if strings.HasPrefix(key.Value, "Key") ||
						strings.HasPrefix(key.Value, "Ctx") {
						fieldname = key.Value
					} else {
						fieldname = fmt.Sprintf("Key%s", key.Value)
					}

					fieldname = strings.ReplaceAll(fieldname, "\"", "")

					if f, ok := fields[fieldname]; ok {
						if "string" != f.Type {
							panic("Conflict duplicate!")
						}
					} else {
						fields[fieldname] = Field{
							Name: fieldname,
							Type: "string",
						}
					}
				}
			}

			return true
		})

		fmt.Println(fields)

		for _, field := range fields {
			name := strings.ToLower(field.Name)
			if field.Type == "error" {
				field.Type = "string"
			}
			if sf, ok := serverFields.Fields[name]; ok {

				if strings.ToLower(sf) != strings.ToLower(field.Type) {
					panic(fmt.Sprintf("filename: %s, type Error: %s on server has %s type, redeclared %s",
						filename,
						field.Name,
						sf,
						field.Type,
					))
				}
			} else {
				toModify.Fields[name] = strings.Title(field.Type)
			}
		}
	}

	if _, err := client.Modify(context.TODO(), &toModify); err != nil {
		panic(err)
	}
}

func workWithRecords(typeSpec *ast.TypeSpec, fields *map[string]Field) {
	//fmt.Println("BBB", typeSpec.Name)

	switch typeSpec.Type.(type) {
	case *ast.StructType:
		structType := typeSpec.Type.(*ast.StructType)
		for _, field := range structType.Fields.List {
			var fieldname string

			if strings.HasPrefix(field.Names[0].Name, "Key") ||
				strings.HasPrefix(field.Names[0].Name, "Ctx") {
				fieldname = field.Names[0].Name
			} else {
				fieldname = fmt.Sprintf("Ctx%s", field.Names[0].Name)
			}

			ftype := fmt.Sprint(field.Type)
			fmt.Println("AAA", field.Type, field.Comment, field.Names)

			if f, ok := (*fields)[fieldname]; ok {
				if ftype != f.Type {
					panic("Conflict duplicate!")
				}
			} else {
				(*fields)[fieldname] = Field{
					Name: fieldname,
					Type: ftype,
				}
			}
		}
	}
}
